import { ValidationError } from 'yup';
import { IAllForms } from './models';
import { EFieldTypes, EnumType, IFormFieldBase, IFormGroupBase, IValidationError, IValidationErrors } from './modelsCommon';
import { IFormParsed } from './modelsParsed';
/**
 * @param jsonSchema - jsonSchema validation object
 * @param t - i18n translate function
 * @returns {any}
 * @description getErrorMessages function returns an object with the error messages for each field in the form based on the jsonSchema
 */
export declare const getErrorMessages: (jsonSchema: any, t?: ((params: any) => void) | undefined) => any;
/**
 * @param translateKey
 * @param param
 * @returns {IValidationError}
 * @description createErrorObject function returns an object with the error message
 * for a field in the form in a proper format
 * @example createErrorObject('required', { param: 'email' }) => { translationKey: 'required', fallbackLanguageTranslation: 'This field is required', meta: { param: 'email' } }
 */
export declare const createErrorObject: (translateKey: string, param?: {
    [key: string]: any;
} | undefined) => IValidationError;
export declare function isFormGroup<T extends EnumType>(toBeDetermined: IFormFieldBase<T> | IFormGroupBase<T>): toBeDetermined is IFormGroupBase<T>;
/**
 * @param stringifiedFunction
 * @returns {Function}
 * @description functionFromString returns a function from a stringified function
 * @example functionFromString('/Function((values) => {return values.medicalConditions.toString() === 'true'})/')
 * => (values) => values.medicalConditions.toString() === 'true'
 */
export declare const functionFromString: (stringifiedFunction: string) => (props: any) => boolean;
/**
 * @param validationError - yup validation error
 * @param initialErrors - initial errors object to be merged with the yup validation errors
 * @returns {IValidationErrors} - returns an object with the errors for each field in the form
 * @description formatYupValidationErrors function returns an object with the errors for each field in the form
 * based on the yup validation error formatting to be used in the GenericForm component
 * @example formatYupValidationErrors(yupValidationError, { email: { property: 'email', constraints: { required: { translationKey: 'required', fallbackLanguageTranslation: 'This field is required', meta: {} } } } })
 */
export declare const formatYupValidationErrors: <T extends EnumType>(validationError: ValidationError, initialErrors?: IValidationErrors<T>) => IValidationErrors<T>;
/**
 * @param children
 * @param type - field type
 * @param acc - accumulator, used for recursion
 * @returns {string[]} - array of field names
 * @description getFieldNamesByType function returns an array of field names based on the field type
 * @example getFieldNamesByType(form, EFieldTypes.TEXT) => ['firstName', 'lastName']
 */
export declare const getFieldNamesByType: <T extends EnumType>(children: (IFormFieldBase<T> | IFormGroupBase<T>)[], type: EFieldTypes, acc?: T[]) => T[];
export declare const flattenValues: <T extends EnumType>(obj: { [key in T]: any; }) => any;
export declare const valuesToDBFormat: <T extends EnumType>(obj: { [key in T]: any; }) => any;
/**
 * @param obj - object to trim
 * @returns {any} - object with trimmed values
 * @description trimValuesInObject function returns an object with trimmed values
 */
export declare const trimValuesInObject: <T extends EnumType>(obj: { [key in T]: any; }) => any;
export declare const parseForm: <T extends EnumType>(form: IAllForms<T>) => IFormParsed<T>;
