import { addYears, isAfter, isBefore, isEqual, isValid, parse, subYears } from 'date-fns';
// @ts-ignore
import { buildYup } from 'json-schema-to-yup';
import get from 'lodash.get';
import isArray from 'lodash.isarray';
import { AnyObjectSchema, ValidationError } from 'yup';
import { EValidationTypes, translatesConfig } from './configs';
import {
  createErrorObject,
  flattenValues,
  formatYupValidationErrors,
  functionFromString,
  getErrorMessages,
  getFieldNamesByType,
  isFormGroup,
  trimValuesInObject,
} from './helpers';
import {
  EFieldTypes,
  EnumType,
  ICrossFieldValidationRule,
  ICrossFieldValidations,
  IFormFieldBase,
  IFormGroupBase,
  IValidationErrorField,
  IValidationErrors,
} from './modelsCommon';
import { IFormParsed } from './modelsParsed';

/**
 * @param formDataChildren - array of form fields
 * @param acc - accumulator, used for recursion
 * @returns {[key: string]: any} - object with the initial values for the form
 * @description getInitialValues function returns an object with the initial values for the form
 * based on the initial values provided in the json schema
 */
export const getInitialValues = <T extends EnumType>(
  formDataChildren: (IFormFieldBase<T> | IFormGroupBase<T>)[],
  acc: any = {},
): { [key: string]: any } => {
  return formDataChildren.reduce((acc, item) => {
    if (isFormGroup(item)) {
      return getInitialValues(item.children, acc);
    }

    acc[item.fieldName] = item.initialValue === null ? undefined : item.initialValue;
    return acc;
  }, acc);
};

/**
 * @param schema - yup schema
 * @param fieldName - field name to get the rule value from
 * @param ruleName - rule name to get the value from
 * @returns {any} - rule value
 * @description getRuleValueFromSchema function returns the value of a rule from a yup schema
 * based on the field name and rule name. This value is needed for passing meta to translations.
 * @example getRuleValueFromSchema(schema, 'email', 'minLength') => 6
 */
export const getRuleValueFromSchema = (
  schema: AnyObjectSchema,
  fieldName: string,
  ruleName: string,
): any => {
  const describedSchema = schema.describe();
  const rules = get(describedSchema, `fields.${fieldName}.tests`, []);
  const rule = rules.find((rule: any) => rule.name === ruleName);

  return get(rule, `params.${ruleName}`, undefined);
};

/**
 * @param type - validation type
 * @param value - value to be validated
 * @param t - translation function(optional)
 * @param validationParam - validation param(meta) to be passed to the translation function(optional) or error object
 * @returns {any} - error object or function, or translated string
 * @description getFieldError function returns an error object or function, or translated string based on the validation type
 */
export const getFieldError = (
  type: string,
  value: string | null,
  t?: (params: any, meta?: any) => void,
  validationParam?: any,
): any => {
  let translationKey = get(translatesConfig, `${type}`);

  switch (type) {
    case EValidationTypes.nullable:
      return t ? t(translationKey) : createErrorObject(translationKey);
    case EValidationTypes.required:
      return t ? t(translationKey) : createErrorObject(translationKey);
    case EValidationTypes.min:
      return t
        ? ({ min }: any) => t(translationKey, { param: min })
        : createErrorObject(translationKey, { param: validationParam });
    case EValidationTypes.max:
      return t
        ? ({ max }: any) => t(translationKey, { param: max })
        : createErrorObject(translationKey, { param: validationParam });
    case EValidationTypes.maxLength:
      return t
        ? ({ maxLength }: any) => t(translationKey, { param: maxLength })
        : createErrorObject(translationKey, { param: validationParam });
    case EValidationTypes.minLength:
      return t
        ? ({ minLength }: any) => t(translationKey, { param: minLength })
        : createErrorObject(translationKey, { param: validationParam });
    case EValidationTypes.format:
      if (value) {
        translationKey = get(translatesConfig, `${type}`)
          ? get(translatesConfig, `${type}`)
          : translationKey;
      }
      return t ? t(translationKey) : createErrorObject(translationKey, { param: validationParam });
    case EValidationTypes.trim:
    case EValidationTypes.oneOf:
    case EValidationTypes.pattern:
      return t ? t(translationKey) : createErrorObject(translationKey, { param: validationParam });
    default:
      return null;
  }
};

/**
 * @param validationObject
 * @param formValues - form values to be validated
 * @returns {IValidationErrors} - object with the validation errors
 * @description makeCrossFieldValidation function makes custom cross-field validation
 * and returns an object with the validation errors
 */
export const makeCrossFieldValidation = <T extends EnumType>(
  validationObject: ICrossFieldValidations<T>,
  formValues: { [key in T]: any },
): IValidationErrors<T> => {
  const errors: IValidationErrors<T> = {};

  try {
    const keys = Object.keys(validationObject) as T[];

    keys.forEach((fieldName) => {
      const rules: ICrossFieldValidationRule[] | undefined = validationObject[fieldName as T];
      const fieldErrors: IValidationErrorField<T> = {
        property: fieldName,
        constraints: {},
      };

      rules?.forEach((rule: any, id: number) => {
        const { errorKey, testCondition, meta } = rule;
        const testFunction = functionFromString(testCondition);
        let shouldIncludeError = false;

        try {
          shouldIncludeError = testFunction(formValues);
        } catch (e) {
          console.log(
            'Error during evaluating custom validation rule. This issue wont cause an error, current validation was skipped',
            e,
          );
        }

        if (shouldIncludeError) {
          fieldErrors.constraints[id] = createErrorObject(errorKey, meta);
        }
      });

      if (Object.keys(fieldErrors.constraints).length > 0) {
        errors[fieldName] = fieldErrors;
      }
    });

    return errors;
  } catch (e) {
    console.warn('Error in makeCrossFieldValidation: ', e);
    return {};
  }
};

/**
 * @param fieldName - field name
 * @param values - form values
 * @param errors - object with the validation errors
 * @returns {IValidationErrors} - object with the validation errors
 * @description validateDateRangeField function validates date range field
 * and returns an object with the validation errors
 */
export const validateDateRangeField = <T extends EnumType>(
  fieldName: T,
  values: any,
  errors: IValidationErrors<T> = {},
): IValidationErrors<T> => {
  const minDate = subYears(new Date(), 100);
  const maxDate = addYears(new Date(), 100);

  const error: IValidationErrorField<T> = {
    property: fieldName,
    constraints: {},
  };

  const dateRangeUnmapped = get(values, fieldName, [null, null]);

  if (!isArray(dateRangeUnmapped)) {
    error.constraints[EValidationTypes.wrongDate] = {
      translationKey: translatesConfig[EValidationTypes.wrongDate],
      fallbackLanguageTranslation: 'This field expected to be array of strings',
      meta: {},
    };

    if (!errors[fieldName]) {
      errors[fieldName] = error;
    } else {
      errors[fieldName]!.constraints = {
        ...errors[fieldName]!.constraints,
        ...error.constraints,
      };
    }

    return errors;
  }

  const dateRange = dateRangeUnmapped.map((value: any) =>
    value ? parse(value, 'yyyy-MM-dd', new Date()) : null,
  );

  if (!isValid(dateRange[0]) || !isValid(dateRange[1])) {
    error.constraints[EValidationTypes.wrongDate] = {
      translationKey: translatesConfig[EValidationTypes.wrongDate],
      fallbackLanguageTranslation: 'Some date in range is invalid',
      meta: {},
    };
  }

  if (!dateRange[0] || !dateRange[1]) {
    error.constraints[EValidationTypes.required] = {
      translationKey: translatesConfig[EValidationTypes.required],
      fallbackLanguageTranslation: 'Some date in range is empty',
      meta: {},
    };
  }

  if (
    (dateRange[0] && isBefore(dateRange[0], minDate)) ||
    (dateRange[1] && isBefore(dateRange[1], minDate))
  ) {
    error.constraints[EValidationTypes.minDate] = {
      translationKey: translatesConfig[EValidationTypes.minDate],
      fallbackLanguageTranslation:
        'Date range should be between 100 years ago and 100 years in the future',
      meta: {},
    };
  }

  if (
    (dateRange[1] && isAfter(dateRange[1], maxDate)) ||
    (dateRange[0] && isAfter(dateRange[0], maxDate))
  ) {
    error.constraints[EValidationTypes.maxDate] = {
      translationKey: translatesConfig[EValidationTypes.maxDate],
      fallbackLanguageTranslation:
        'Date range should be between 100 years ago and 100 years in the future',
      meta: {},
    };
  }

  if (
    dateRange[0] &&
    dateRange[1] &&
    (isAfter(dateRange[0], dateRange[1]) || isEqual(dateRange[0], dateRange[1]))
  ) {
    error.constraints[EValidationTypes.wrongRange] = {
      translationKey: translatesConfig[EValidationTypes.wrongRange],
      fallbackLanguageTranslation: 'Start date should be before end date',
      meta: {},
    };
  }

  if (Object.keys(error.constraints).length > 0) {
    if (!errors[fieldName]) {
      errors[fieldName] = error;
    } else {
      errors[fieldName]!.constraints = {
        ...errors[fieldName]!.constraints,
        ...error.constraints,
      };
    }
  }

  return errors;
};

/**
 * @param jsonSchema - json validation schema
 * @param t - translation function(optional)
 * @returns {AnyObjectSchema} - yup validation schema
 * @description buildYupSchema function builds a yup validation schema based on the json validation schema
 */
export const buildYupSchema = (jsonSchema: any, t?: (params: any) => void): AnyObjectSchema => {
  return buildYup(
    jsonSchema,
    t
      ? {
          errMessages: getErrorMessages(jsonSchema, t),
        }
      : undefined,
  );
};

/**
 * @param form {IForm} - form json schema
 * @param valuesToValidate {any} - form values to be validated
 * @param skipCrossField {boolean} - skip cross-field validation(optional)
 * @returns {Promise<unknown>} - object with the validation errors or error that has occurred during the validation
 * @description validateValues function validates the form values based on the json form object
 */
export const validateValues = async <T extends EnumType>(
  { children, jsonSchema, formSettings }: IFormParsed<T>,
  valuesToValidate: { [key in T]: any },
  skipCrossField: boolean = false,
): Promise<unknown> => {
  let errors: any = {};
  const flattenedValues = flattenValues(valuesToValidate);
  const trimmedValues = trimValuesInObject(flattenedValues);

  try {
    const validationSchema = buildYupSchema(jsonSchema);
    const dateRangeFields = getFieldNamesByType(children, EFieldTypes.DATE_RANGE);

    if (!skipCrossField) {
      const formFieldNames = Object.keys(formSettings) as T[];
      const crossFieldValidations = formFieldNames.reduce((acc, el) => {
        if (!formSettings[el].crossFieldValidations) {
          return acc;
        }

        acc[el] = formSettings[el].crossFieldValidations;
        return acc;
      }, {} as ICrossFieldValidations<T>);

      errors = makeCrossFieldValidation(crossFieldValidations, trimmedValues);
    }

    if (dateRangeFields.length > 0) {
      dateRangeFields.forEach((fieldName) => {
        errors = validateDateRangeField(fieldName, trimmedValues, errors);
      });
    }

    await validationSchema.validate(trimmedValues, { abortEarly: false });
  } catch (e) {
    if (get(e, 'name') === 'ValidationError') {
      errors = formatYupValidationErrors(e as ValidationError, errors);
    } else {
      throw e;
    }
  }
  return { errors: Object.values(errors) };
};
